from typing import List
from typing import Union


def new_mixdb_from_mixid(mixdb: dict,
                         mixid: Union[str, List[int]]) -> dict:
    from copy import deepcopy

    from sonusai import SonusAIError
    from sonusai.mixture.get_mixtures_from_mixid import get_mixtures_from_mixid

    mixdb_out = deepcopy(mixdb)
    mixdb_out['mixtures'] = get_mixtures_from_mixid(mixdb_out, mixid)

    if not mixdb_out['mixtures']:
        raise SonusAIError(f'Error processing mixid: {mixid}; resulted in empty list of mixtures')

    return mixdb_out



from typing import List
from typing import Union

import numpy as np


def strictly_decreasing(list_to_check: list) -> bool:
    return all(x > y for x, y in zip(list_to_check, list_to_check[1:]))


def generate_truth(mixdb: dict,
                   record: dict,
                   target: np.ndarray,
                   noise: np.ndarray,
                   compute: bool = True) -> np.ndarray:
    from copy import deepcopy

    from sonusai.mixture.class_count import get_class_count
    from sonusai.mixture.get_class_weights_threshold import get_class_weights_threshold

    if not compute:
        return np.empty(0, dtype=np.single)

    truth = np.zeros((len(target), mixdb['num_classes']), dtype=np.single)
    for truth_setting in mixdb['targets'][record['target_file_index']]['truth_settings']:
        config = deepcopy(truth_setting)
        config['frame_size'] = mixdb['frame_size']
        config['num_classes'] = mixdb['num_classes']
        config['mutex'] = mixdb['truth_mutex']
        config['target_gain'] = record['target_gain']
        new_truth = truth_function(target=target,
                                   noise=noise,
                                   config=config)
        truth = truth + new_truth

    indices = [sub['index'] for sub in mixdb['targets'][record['target_file_index']]['truth_settings']]
    record['class_count'] = get_class_count(
        truth_index=indices,
        truth=truth,
        class_weights_threshold=get_class_weights_threshold(mixdb))

    return truth


def truth_function(target: np.ndarray,
                   noise: np.ndarray,
                   config: dict) -> np.ndarray:
    import h5py
    from pyaaware import ForwardTransform
    from pyaaware import SED

    from sonusai import SonusAIError
    from sonusai.utils import int16_to_float

    if config['function'] == 'sed':
        if config['target_gain'] == 0:
            return np.zeros((len(target), config['num_classes']), dtype=np.single)

        if len(target) % config['frame_size'] != 0:
            raise SonusAIError(f'Number of samples in audio is not a multiple of {config["frame_size"]}')

        if 'config' not in config:
            raise SonusAIError('Truth function SED missing config')

        parameters = ['thresholds']
        for parameter in parameters:
            if parameter not in config['config']:
                raise SonusAIError(f'Truth function SED config missing required parameter: {parameter}')

        thresholds = config['config']['thresholds']
        if not strictly_decreasing(thresholds):
            raise SonusAIError(f'Truth function SED thresholds are not strictly decreasing: {thresholds}')

        fft = ForwardTransform(N=config['frame_size'] * 4, R=config['frame_size'])
        sed = SED(thresholds=thresholds,
                  index=config['index'],
                  frame_size=config['frame_size'],
                  num_classes=config['num_classes'],
                  mutex=config['mutex'])

        audio = np.int16(np.single(target) / config['target_gain'])
        truth = np.empty((0, config['num_classes']), dtype=np.single)
        for offset in range(0, len(audio), config['frame_size']):
            new_truth = sed.execute(fft.energy(int16_to_float(audio[offset:offset + config['frame_size']])))
            truth = np.vstack((truth, np.reshape(new_truth, (1, len(new_truth)))))

        truth = truth.repeat(config['frame_size'], axis=0)
        return truth

    elif config['function'] == 'file':
        if 'config' not in config:
            raise SonusAIError('Truth function file missing config')

        parameters = ['file']
        for parameter in parameters:
            if parameter not in config['config']:
                raise SonusAIError(f'Truth function file config missing required parameter: {parameter}')

        with h5py.File(name=config['config']['file'], mode='r') as f:
            truth_in = f['/truth_t'][:]

        if truth_in.ndim != 2:
            raise SonusAIError('Truth file data is not 2 dimensions')

        if truth_in.shape[0] != len(target):
            raise SonusAIError('Truth file does not contain the right amount of samples')

        truth = np.zeros((len(target), config['num_classes']), dtype=np.single)
        if config['target_gain'] == 0:
            return truth

        if isinstance(config['index'], list):
            if len(config['index']) != truth_in.shape[1]:
                raise SonusAIError('Truth file does not contain the right amount of classes')

            truth[:, config['index']] = truth_in
        else:
            if config['index'] + truth_in.shape[1] > config['num_classes']:
                raise SonusAIError('Truth file contains too many classes')

            truth[:, config['index']:config['index'] + truth_in.shape[1]] = truth_in

        return truth

    elif config['function'] == 'snr':
        truth = np.zeros((len(target), config['num_classes']), dtype=np.single)
        fft = ForwardTransform(N=config['frame_size'] * 4, R=config['frame_size'])
        for offset in range(0, len(target), config['frame_size']):
            target_energy = fft.energy(int16_to_float(target[offset:offset + config['frame_size']]))
            noise_energy = fft.energy(int16_to_float(noise[offset:offset + config['frame_size']]))
            truth[offset:offset + config['frame_size'], config['index']] = np.single(target_energy / noise_energy)

        return truth

    elif config['function'] == 'phoneme':
        # Read in .txt transcript and run a Python function to generate text grid data
        # (indicating which phonemes are active)
        # Then generate truth based on this data and put in the correct classes based on config['index']
        raise SonusAIError('Truth function phoneme is not supported yet')

    raise SonusAIError(f'Unsupported truth function: {config["function"]}')


def get_truth_indices(mixdb: dict, mixid: Union[str, List[int]]) -> list:
    from sonusai.mixture.get_mixtures_from_mixid import convert_mixid_to_list

    indices = list()

    for m in convert_mixid_to_list(mixdb, mixid):
        target_file_index = mixdb['mixtures'][m]['target_file_index']

        # extract indices from truth_settings
        index = [sub['index'] for sub in mixdb['targets'][target_file_index]['truth_settings']]

        # flatten, uniquify, and sort
        index = sorted(list(set([item for sublist in index for item in sublist])))

        indices.append(index)

    return indices



import warnings

import numpy as np


def one_hot(truth: np.ndarray,
            predict: np.ndarray,
            predict_thr: float = 0,
            truth_thr: float = 0.5,
            timesteps: int = -1) -> (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray):
    """ Calculates metrics from one-hot prediction and truth data (numpy float arrays) where
        both are one-hot probabilities (or quantized decisions) for each class
        with size frames x num_classes or frames x timesteps x num_classes.
        For metrics that require it, truth and pred decisions will be made using threshold >= pthr.
        Some metrics like AP and AUC do not depend on pthr for predict, but still use truth >= pthr

        pred_thr sets the decision threshold(s) applied to predict data for some metrics, thus allowing
        the input to be continuous probabilities, for AUC-type metrics and root mean square error (rmse).
        1. Default = 0 (multiclass or binary) which infers:
           binary (nclass = 1)         use >= 0.5 for truth and pred (same as argmax() for binary)
           multi-class/single-label    if truth_mutex= = true, use argmax() used on both truth and pred
           note multi-label metrics are disabled for pthr = 0, must set pthr > 0

        2. pthr > 0 (multilabel or binary) scalar or a vector num_class x 1 then use
           pthr as a binary decision threshold in each class:
           binary (nclass = 1)         use >= pthr[0] for pred and pthr[num_classes+1] for truth
                                       if it exists, else use >= 0.5 for truth
           multi-label                 use >= pthr for pred if scalar, or pthr[class_idx] if vector
                                       use >= pthr[num_classes+1] for truth if exists, else 0.5
           note multi-class/single-label inputs are meaningless in this mode, use pthr = 0 argmax mode

        num_classes is inferred from 1D, 2D or 3D truth inputs by default (default tstep = -1 or 0).
        Only set tstep > 0 in case of ambiguous binary 2D case where input frames x timesteps,
        then it must set to the number of timesteps = TSTEPS (which will be > 0).
        It is safe to always set tstep <= 0 for binary inputs, and if truth_shape[2] exists

        returns metrics over all frames + timesteps:
        mcm      num_classes x 2 x 2         multiclass confusion matrix count ove
        metrics  num_classes x 14            [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP, AP, AUC]
        cm, cmn: num_classes x num_classes   confusion matrix, normalized confusion matrix
        rmse:    num_classes x 1             RMS error over all frames + timesteps, before threshold decision
        mavg:    3x8 averages                macro,micro,weighted [PPV, TPR, F1, FPR, ACC, mAP, mAUC, TPSUM]
    """
    from sklearn.metrics import average_precision_score
    from sklearn.metrics import confusion_matrix
    from sklearn.metrics import multilabel_confusion_matrix
    from sklearn.metrics import precision_recall_fscore_support
    from sklearn.metrics import roc_auc_score

    from sonusai import logger

    if truth.shape != predict.shape:
        logger.error('Shape of truth and predict are not equal')
        exit()

    # truth, predict can be either frames x num_classes, or frames x timesteps x num_classes
    # in binary case dim may not exist, detect this and set num_classes == 1
    if truth.ndim == 3 or (truth.ndim == 2 and timesteps > 0):
        if truth.ndim == 2:  # 2D but has timesteps = truth.shape[1]
            num_classes = 1
        else:
            num_classes = truth.shape[2]  # frames = truth.shape[0], timesteps = truth.shape[1]
        # reshape to remove timestep dimension
        truth = np.reshape(truth, (truth.shape[0] * truth.shape[1], truth.shape[2]))
        predict = np.reshape(predict, (predict.shape[0] * predict.shape[1], predict.shape[2]))
    else:
        if truth.ndim == 1:  # no timesteps dimension, = 0
            num_classes = 1
            truth = np.expand_dims(truth, 1)  # convert to 2D (F,1) if only 1 dim
        else:
            num_classes = truth.shape[1]  # 2D input frames x num_classes

    # Support predict (and truth) when shape is (F,), just convert to (F,1)
    if predict.ndim == 1:
        predict = np.expand_dims(predict, 1)

    # Regression metric root-mean-square-error always works
    rmse = np.sqrt(np.mean(np.square(truth - predict), axis=0))

    # Calc default predict decision thresholds based on mode
    if not np.ndim(predict_thr) and predict_thr == 0:  # if scalar and 0, set defaults
        if num_classes == 1:
            predict_thr = np.atleast_1d(0.5)  # binary case default >= 0.5 which is equiv to argmax()
        else:
            predict_thr = np.atleast_1d(0)  # multiclass, single-label (argmax mode)
    else:
        if np.ndim(predict_thr) > 1:
            assert predict_thr.shape[0] == num_classes  # multi-label with custom thr vector
        else:
            if predict_thr == 0:
                predict_thr = np.atleast_1d(0.5)  # binary or multi-label scalar default
            else:
                predict_thr = np.atleast_1d(predict_thr)  # user specified binary or multi-label scalar

    # if predict_thr[0] == 0 and num_classes > 1:  # if scalar
    #     pthr = 0.5  # multiclass, single-label (argmax) or multilabel case default
    # else:
    #     if num_classes == 1 and predict_thr[0] == 0:
    #         pthr = 0.5  # binary case default >= 0.5 which is equiv to argmax()
    #     else:
    #         pthr = pthr  # any case using specified threshold

    # Convert continuous probabilities to binary via argmax() or threshold comparison
    # and create labels of int encoded (0:num_classes-1), and then equivalent one-hot
    if num_classes == 1:  # If binary
        binary_mode = True
        labels = ([i for i in range(0, 2)])  # int encoded 0,1
        plabel = np.int8(predict >= predict_thr)  # frames x 1, default 0.5 is equiv. to argmax()
        tlabel = np.int8(truth >= truth_thr)  # frames x 1
        predb = plabel
        truthb = tlabel
    else:
        binary_mode = False
        labels = ([i for i in range(0, num_classes)])  # int encoded 0,...,num_classes-1
        if predict_thr[0] == 0:  # multiclass single-label (mutex), use argmax
            plabel = np.argmax(predict, axis=-1)  # frames x 1 labels
            tlabel = np.argmax(truth, axis=-1)  # frames x 1 labels
            predb = np.zeros(predict.shape, dtype=np.int8)  # frames x num_classes one-hot binary
            truthb = np.zeros(truth.shape, dtype=np.int8)  # frames x num_classes one-hot binary
            predb[np.arange(predb.shape[0]), plabel] = 1  # single-label frames x num_classes
            truthb[np.arange(truthb.shape[0]), tlabel] = 1  # single-label frames x num_classes
        else:  # multi-label prob threshold comparison (multiple classes)
            predb = np.int8(predict >= predict_thr.transpose())  # frames x num_classes multilabel one-hot decision
            truthb = np.int8(truth >= truth_thr)  # frames x num_classes multilabel one-hot decision
            # Return argmax() for optional single-label confusion matrix metrics
            plabel = np.argmax(predict, axis=-1)  # frames x 1 labels
            tlabel = np.argmax(truth, axis=-1)  # frames x 1 labels

    # debug checks to understand ap, auc:
    # from sklearn.metrics import roc_curve
    # fpr, tpr, thr = roc_curve(truthb[:,0],predict[:,0],drop_intermediate=False)
    # from sklearn.metrics import precision_recall_curve
    # precision, recall, thr  = precision_recall_curve(truthb[:,0], predict[:,0])
    # from sklearn.metrics import RocCurveDisplay
    # RocCurveDisplay.from_predictions(truthb[:,0],predict[:,0])  # Plot ROC class0

    # Create nclass x 2 x 2 multi-label confusion matrix (mcm)
    # Note - must include labels or sklearn func. will omit non-exiting classes
    mcm = multilabel_confusion_matrix(truthb, predb, labels=labels)

    if num_classes == 1:
        mcm = mcm[1:]  # remove dim 0 if binary

    # Create num_classes x num_classes normalized confusion matrix
    cmn = confusion_matrix(tlabel, plabel, labels=labels, normalize='true')

    # Create num_classes x num_classes confusion matrix
    cm = confusion_matrix(tlabel, plabel, labels=labels)

    # Combine all per-class metrics into a single array
    # [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP, AP, AUC]
    metrics = np.zeros((num_classes, 14))
    # threshold_optpr = np.zeros((num_classes, 1))
    eps = np.finfo(float).eps
    for nci in range(num_classes):
        # True negative
        TN = mcm[nci, 0, 0]
        # False positive
        FP = mcm[nci, 0, 1]
        # False negative
        FN = mcm[nci, 1, 0]
        # True positive
        TP = mcm[nci, 1, 1]
        # Accuracy
        ACC = (TP + TN) / (TP + TN + FP + FN + eps)
        # True positive rate, sensitivity, recall, hit rate (note eps in numerator)
        # When ``true positive + false negative == 0``, recall is undefined, set to 0
        TPR = (TP) / (TP + FN + eps)
        # Precision, positive predictive value
        # When ``true positive + false positive == 0``, precision is undefined, set to 0
        PPV = TP / (TP + FP + eps)
        # Specificity i.e., selectivity, or true negative rate
        TNR = TN / (TN + FP + eps)
        # False positive rate = 1-specificity, roc x-axis
        FPR = FP / (TN + FP + eps)
        # HitFA used by some separation research, close match to MCC
        HITFA = TPR - FPR
        # F1 harmonic mean of precision, recall = 2*PPV*TPR / (PPV + TPR)
        F1 = 2 * TP / (2 * TP + FP + FN + eps)
        # Matthew correlation coefficient
        MCC = (TP * TN - FP * FN) / (np.sqrt((TP + FP) * (TP + FN) * (TN + FP) * (TN + FN)) + eps)
        # Num. negatives total (truth), also = TN+FP denom of FPR
        NT = sum(mcm[nci, 0,])
        # Num. positives total (truth), also = FN+TP denom of TPR, precision
        PT = sum(mcm[nci, 1,])
        # Average Precision also called area under the PR curve AUCPR and
        # AUC ROC curve using binary-ized truth and continuous prediction probabilities
        # sklearn returns nan if no active truth in a class but w/un-suppressible div-by-zero warning
        if np.sum(truthb[:, nci]) == 0:  # no active truth must be NaN
            AUC = np.NaN
            AP = np.NaN
            # threshold_optpr[nci] = np.NaN
        else:
            AP = average_precision_score(truthb[:, nci], predict[:, nci], average=None)
            AUC = roc_auc_score(truthb[:, nci], predict[:, nci], average=None)
            # # Optimal threshold from PR curve, optimizes f-score
            # precision, recall, thresholds = precision_recall_curve(truthb[:, nci], predict[:, nci])
            # fscore = (2 * precision * recall) / (precision + recall)
            # ix = argmax(fscore)  # index of largest f1 score
            # threshold_optpr[nci] = thresholds[ix]

        metrics[nci] = [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP, AP, AUC]

    # Calculate averages into single array, 3 types for now Macro, Micro, Weighted
    mavg = np.zeros((3, 8), dtype=np.single)
    s = np.sum(metrics[:, 9].astype(int))  # support = sum (true pos total = FN+TP ) over classes

    # macro average [PPV, TPR, F1, FPR, ACC, mAP, mAUC, TPSUM]
    with warnings.catch_warnings():
        warnings.filterwarnings(action='ignore', message='Mean of empty slice')
        mavg[0,] = [np.mean(metrics[:, 2]), np.mean(metrics[:, 1]), np.mean(metrics[:, 6]), np.mean(metrics[:, 4]),
                    np.mean(metrics[:, 0]), np.nanmean(metrics[:, 12]), np.nanmean(metrics[:, 13]), s]

    # micro average, micro-F1 = micro-precision = micro-recall = accuracy
    if num_classes > 1:
        tp_sum = np.sum(metrics[:, 10])  # TP all classes
        rm = tp_sum / (np.sum(metrics[:, 9]) + eps)  # micro mean PPV = TP / (PT=FN+TP)
        fp_sum = np.sum(metrics[:, 11])  # FP false-positives all classes
        fpm = fp_sum / (np.sum(metrics[:, 8]) + eps)  # micro mean FPR = FP / (NT=TN+FP)
        pm = tp_sum / (tp_sum + fp_sum + eps)  # micro mean TPR = TP / (TP+FP) (note: same as rm for micro-avg)
        fn_sum = sum(mcm[:, 1, 0])
        f1m = 2 * tp_sum / (2 * tp_sum + fp_sum + fn_sum + eps)
        tn_sum = sum(mcm[:, 0, 0])
        accm = (tp_sum + tn_sum) / (tp_sum + tn_sum + fp_sum + fn_sum + eps)
        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore', message='invalid value encountered in true_divide')
            miap = average_precision_score(truthb, predict, average='micro')
        if np.sum(truthb):  # no activity over all classes
            miauc = roc_auc_score(truthb, predict, average='micro')
        else:
            miauc = np.NaN

        # [miPPV, miTPR, miF1, miFPR, miACC, miAP, miAUC, TPSUM]
        mavg[1,] = [pm, rm, f1m, fpm, accm, miap, miauc, s]  # specific format, last 3 are unique

        # weighted average TBD
        wp, wr, wf1, _ = precision_recall_fscore_support(truthb, predb, average='weighted', zero_division=0)
        if np.sum(truthb):
            taidx = np.sum(truthb, axis=0) > 0
            wap = average_precision_score(truthb[:, taidx], predict[:, taidx], average='weighted')
            wauc = roc_auc_score(truthb[:, taidx], predict[:, taidx], average='weighted')
        else:
            wap = np.NaN
            wauc = np.NaN

        mavg[2, :] = [wp, wr, wf1, 0, 0, wap, wauc, s]
    else:  # binary case, all are same
        mavg[1,] = mavg[0,]
        mavg[2,] = mavg[0,]

    return mcm, metrics, cm, cmn, rmse, mavg



import warnings
from typing import List
from typing import Union


def snr_summary(mixdb: dict,
                mixid: Union[str, List[int]],
                truth_f,
                predict,
                segsnr=[],
                pred_thr=0,
                truth_thr=0.5,
                timesteps=0):
    """ Calculate average-over-class metrics per SNR over specified mixture list.
        Inputs:
          mixdb        mixture database (Sonusai .json format)
          mixid
          truth_f      truth/labels      #features x num_classes
          predict      prediction data / neural net model one-hot out,  #features x num_classes
          segsnr       segmental snr from Sonusai genft,  #transform frames x 1
          pred_thr     decision threshold(s) applied to predict data, allowing predict to be
                       continuous probabilities or decisions

        Default pred_thr=0 will infer 0.5 for multi-label mode (truth_mutex = False), or
        if single-label mode (truth_mutex == True) then ignore and use argmax mode, and
        the confusion matrix is calculated for all classes.

        Returns pandas dataframe (snrdf) of metrics per snr.

        Example:

    """
    import numpy as np
    import pandas as pd

    from sonusai.metrics.one_hot import one_hot
    from sonusai.mixture import get_mixid_data
    from sonusai.queries.get_organized_snrs import get_mixids_from_ordered_snr

    num_classes = truth_f.shape[1]

    mxid_snro = get_mixids_from_ordered_snr(mixdb=mixdb, mixid=mixid)

    # Check pred_thr array or scalar and return final scalar pred_thr value
    if not mixdb['truth_mutex'] and num_classes > 1:
        if np.ndim(pred_thr) == 0 and pred_thr == 0:
            pred_thr = 0.5  # multi-label pred_thr scalar 0 force to 0.5 default

        if np.ndim(pred_thr) == 1:
            if len(pred_thr) == 1:
                if pred_thr[0] == 0:
                    pred_thr = 0.5  # multi-label pred_thr array scalar 0 force to 0.5 default
                else:
                    pred_thr = pred_thr[0]  # multi-label pred_thr array set to scalar = array[0]

    if len(segsnr) > 0:  # prep segsnr if provided, transform frames to feature frames via mean()
        NFF = int(segsnr.shape[0] / truth_f.shape[0])  # expected to always be an integer
        segsnr_f = np.mean(np.reshape(segsnr, (truth_f.shape[0], NFF)), axis=1, keepdims=True)
        ssnr_stats = np.zeros((len(mxid_snro), 3), dtype=float)

    macro_avg = np.zeros((len(mxid_snro), 7), dtype=float)
    micro_avg = np.zeros((len(mxid_snro), 7), dtype=float)
    wghtd_avg = np.zeros((len(mxid_snro), 7), dtype=float)
    ii = 0
    for snrkey in mxid_snro:
        ytrue, ypred = get_mixid_data(mixdb, mxid_snro[snrkey], truth_f, predict)
        if mixdb['truth_mutex']:
            _, metrics, _, _, _, mavg = one_hot(ytrue[:, 0:-1], ypred[:, 0:-1], pred_thr, truth_thr, timesteps)
        else:
            _, metrics, _, _, _, mavg = one_hot(ytrue, ypred, pred_thr, truth_thr, timesteps)
        # mavg macro, micro, weighted: [PPV, TPR, F1, FPR, ACC, mAP, mAUC, TPSUM]
        macro_avg[ii, :] = mavg[0, 0:7]
        micro_avg[ii, :] = mavg[1, 0:7]
        wghtd_avg[ii, :] = mavg[2, 0:7]
        if len(segsnr) > 0:
            ytrue, ysegsnr = get_mixid_data(mixdb, mxid_snro[snrkey], truth_f, segsnr_f)
            with warnings.catch_warnings():
                warnings.filterwarnings(action='ignore', message='divide by zero encountered in log10')
                # segmental SNR mean = mixture_snr and target_snr
                ssnr_stats[ii, 0] = 10 * np.log10(np.mean(ysegsnr))
                # seg SNR 80% percentile
                ssnr_stats[ii, 1] = 10 * np.log10(np.percentile(ysegsnr, 80, interpolation='midpoint'))
                # seg SNR max
                ssnr_stats[ii, 2] = 10 * np.log10(max(ysegsnr))

        ii = ii + 1

    # SNR format SNR: PPV, TPR, F1, FPR, ACC, AP, AUC, SNRSEG80P
    col_n = ['PPV', 'TPR', 'F1', 'FPR', 'ACC', 'AP', 'AUC']
    snr_macrodf = pd.DataFrame(macro_avg, index=list(mxid_snro.keys()), columns=col_n)
    snr_macrodf.sort_index(ascending=False, inplace=True)

    snr_microdf = pd.DataFrame(micro_avg, index=list(mxid_snro.keys()), columns=col_n)
    snr_microdf.sort_index(ascending=False, inplace=True)

    snr_wghtdf = pd.DataFrame(wghtd_avg, index=list(mxid_snro.keys()), columns=col_n)
    snr_wghtdf.sort_index(ascending=False, inplace=True)

    # Add segmental snr columns if provided
    if len(segsnr) > 0:
        ssnrdf = pd.DataFrame(ssnr_stats, index=list(mxid_snro.keys()), columns=['SSNRavg', 'SSNR80p', 'SSNRmax'])
        ssnrdf.sort_index(ascending=False, inplace=True)
        snr_macrodf = pd.concat([snr_macrodf, ssnrdf], axis=1)
        snr_microdf = pd.concat([snr_microdf, ssnrdf], axis=1)
        snr_wghtdf = pd.concat([snr_wghtdf, ssnrdf], axis=1)

    return snr_macrodf, snr_microdf, snr_wghtdf, mxid_snro



import numpy as np

from sonusai.mixture import get_file_frame_segments


def get_mixid_data(mixdb, mixid, truth_f, predict):
    """
    Collect per-feature data of specified mixids from mixdb where inputs are:
       truth_f:   truth data matching mixdb (size #feature_frames x num_classes)
       predict:   prediction or segsnr data size #feature_frames x ndim (ndim > 1)

    Returns:
        ytrue:    np.array combined truth from mixids
        ypred:    np.array combined data from mixids
    """
    num_classes = truth_f.shape[1]
    dnum = predict.shape[1]  # same as num_class for prediction data, but use for segsnr too
    file_frame_segments = get_file_frame_segments(mixdb, mixid)
    total_frames = sum([file_frame_segments[m].length for m in file_frame_segments])
    ytrue = np.empty((total_frames, num_classes), dtype=np.single)
    ypred = np.empty((total_frames, dnum), dtype=np.single)
    start = 0
    for m in file_frame_segments:
        length = file_frame_segments[m].length
        ytrue[start:start + length] = truth_f[file_frame_segments[m].get_slice()]
        ypred[start:start + length] = predict[file_frame_segments[m].get_slice()]
        start += length

    return ytrue, ypred



def update_truth_settings(truth_settings: list, default: list = None) -> None:
    from sonusai import SonusAIError

    if default is not None and len(truth_settings) != len(default):
        raise SonusAIError(f'Length of truth_settings does not match given default')

    required_keys = [
        'function',
        'config',
        'index',
    ]
    for n in range(len(truth_settings)):
        for key in required_keys:
            if key not in truth_settings[n]:
                if default is not None and key in default[n]:
                    truth_settings[n][key] = default[n][key]
                else:
                    raise SonusAIError(f"Missing required '{key}' in truth_settings")



def get_config_from_file(config_file: str) -> dict:
    from sonusai.mixture.get_default_config import get_default_config
    from sonusai.mixture.update_config_from_file import update_config_from_file

    return update_config_from_file(name=config_file, config=get_default_config())



def get_default_config() -> dict:
    import sonusai
    from sonusai.mixture.load_config import load_config

    try:
        return load_config(sonusai.mixture.DEFAULT_CONFIG)
    except Exception as e:
        raise sonusai.SonusAIError(f'Error loading default config: {e}')



def load_config(name: str) -> dict:
    import yaml

    with open(file=name, mode='r') as f:
        config = yaml.safe_load(f)

    config_variable_substitution(config)
    return config


def config_variable_substitution(config: dict) -> None:
    """Find custom SonusAI variables in given dictionary and substitute their values in place"""
    import yaml

    import sonusai

    for key, value in config.items():
        print(f'config[{key}] = {value}')
        if isinstance(value, dict):
            config_variable_substitution(value)
        else:
            if isinstance(value, str):
                if value == '${default_noise}':
                    config[key] = sonusai.mixture.DEFAULT_NOISE
                    return
                if value == '${frame_size}':
                    config[key] = sonusai.mixture.DEFAULT_FRAME_SIZE
                    return



import numpy as np


def generate_segsnr(mixdb: dict,
                    record: dict,
                    target: np.ndarray,
                    noise: np.ndarray,
                    compute: bool = True,
                    frame_based: bool = False) -> np.ndarray:
    from pyaaware import ForwardTransform

    from sonusai.utils import int16_to_float

    if not compute:
        return np.empty(0, dtype=np.single)

    fft = ForwardTransform(N=mixdb['frame_size'] * 4, R=mixdb['frame_size'])

    if frame_based:
        segsnr = np.empty(record['samples'] // mixdb['frame_size'], dtype=np.single)
    else:
        segsnr = np.empty(record['samples'], dtype=np.single)

    frame = 0
    for offset in range(0, record['samples'], mixdb['frame_size']):
        indices = slice(offset, offset + mixdb['frame_size'])

        target_energy = fft.energy(int16_to_float(target[indices]))
        noise_energy = fft.energy(int16_to_float(noise[indices]))

        if noise_energy == 0:
            snr = np.single(np.inf)
        else:
            snr = np.single(target_energy / noise_energy)

        if frame_based:
            segsnr[frame] = snr
            frame += 1
        else:
            segsnr[indices] = snr

    return segsnr



from typing import List
from typing import Union


def confusion_matrix_summary(mixdb: dict,
                             mixid: Union[str, List[int]],
                             truth_f,
                             predict,
                             class_idx,
                             pred_thr=0,
                             truth_thr=0.5,
                             timesteps=0):
    """ Calculate confusion matrix for specified class, using truth and prediction
        data features x num_classes.

        pred_thr sets the decision threshold(s) applied to predict data, thus allowing
        predict to be continuous probabilities.

        Default pred_thr=0 will infer 0.5 for multi-label mode (truth_mutex = False), or
        if single-label mode (truth_mutex == True) then ignore and use argmax mode, and
        the confusion matrix is calculated for all classes.

        Returns pandas dataframes of confusion matrix cmdf and normalized confusion matrix cmndf.

        Example:

    """
    import numpy as np
    import pandas as pd

    from sonusai.metrics.one_hot import one_hot
    from sonusai.mixture import get_file_frame_segments

    num_classes = truth_f.shape[1]

    file_frame_segments = get_file_frame_segments(mixdb, mixid)
    total_frames = sum([file_frame_segments[m].length for m in file_frame_segments])
    ytrue = np.empty((total_frames, num_classes), dtype=np.single)
    ypred = np.empty((total_frames, num_classes), dtype=np.single)
    start = 0
    for m in file_frame_segments:
        length = file_frame_segments[m].length
        ytrue[start:start + length] = truth_f[file_frame_segments[m].get_slice()]
        ypred[start:start + length] = predict[file_frame_segments[m].get_slice()]
        start += length

    # Check pred_thr array or scalar and return final scalar pred_thr value
    if not mixdb['truth_mutex'] and num_classes > 1:
        if np.ndim(pred_thr) == 0 and pred_thr == 0:
            pred_thr = 0.5  # multi-label pred_thr scalar 0 force to 0.5 default

        if np.ndim(pred_thr) == 1:
            if len(pred_thr) == 1:
                if pred_thr[0] == 0:
                    pred_thr = 0.5  # multi-label pred_thr array scalar 0 force to 0.5 default
                else:
                    pred_thr = pred_thr[0]  # multi-label pred_thr array set to scalar = array[0]
            else:
                pred_thr = pred_thr[class_idx]  # multi-label pred_thr array scalar set = array[class_idx]

    if len(mixdb['class_labels']) == num_classes:
        class_names = mixdb['class_labels']
    else:
        class_names = ([f'Class {i}' for i in range(1, num_classes + 1)])

    class_nums = ([f'{i}' for i in range(1, num_classes + 1)])

    if mixdb['truth_mutex']:
        pred_thr = 0  # single-label mode force to argmax mode
        _, _, cm, cmn, _, _ = one_hot(ytrue, ypred, pred_thr, truth_thr, timesteps)
        row_n = class_names
        row_n[-1] = 'Other'
        # mux = pd.MultiIndex.from_product([['Single-label/mutex mode, truth thr = {}'.format(truth_thr)],
        #                                   class_nums])
        # mux = pd.MultiIndex.from_product([['truth thr = {}'.format(truth_thr)], class_nums])

        cmdf = pd.DataFrame(cm, index=row_n, columns=class_nums, dtype=int)
        cmndf = pd.DataFrame(cmn, index=row_n, columns=class_nums, dtype=float)

    else:
        _, _, cm, cmn, _, _ = one_hot(ytrue[:, class_idx], ypred[:, class_idx], pred_thr, truth_thr, timesteps)
        cname = class_names[class_idx]
        row_n = ['TrueN', 'TrueP']
        col_n = ['N-' + cname, 'P-' + cname]
        cmdf = pd.DataFrame(cm, index=row_n, columns=col_n, dtype=int)
        cmndf = pd.DataFrame(cmn, index=row_n, columns=col_n, dtype=float)
        # add thresholds in 3rd row
        pdnote = pd.DataFrame(np.atleast_2d([pred_thr, truth_thr]), index=['p/t thr:'], columns=col_n)
        cmdf = pd.concat([cmdf, pdnote])
        cmndf = pd.concat([cmndf, pdnote])

    return cmdf, cmndf



"""sonusai evaluate

usage: evaluate [-hv] (-f FEATURE) (-n PREDICT) [-t PTHR] [-p PLOTNUM]

options:
   -h, --help
   -v, --verbose                    Be verbose.
   -f FEATURE, --feature FEATURE    Feature + truth .h5 data file
   -n PREDICT, --predict PREDICT    Predict .h5 data file.
   -t PTHR, --thr PTHR              Optional prediction decision threshold, 0 = argmax(). [default: 0].
   -p PLOTNUM, --plotnum PLOTNUM    Optional plot mixture results (-1 is plot all, 0 is plot none) [default: 0].

Evaluate calculates performance metrics of neural-network models from model prediction data and the associated
feature+truth data file.

Inputs:
    FEATURE     A SonusAI feature+truth HDF5 file. Contains:
                    attribute:  mixdb
                    dataset:    feature   Note: not required for metrics
                    dataset:    truth_f
                    dataset:    segsnr    Optional
    PREDICT     A SonusAI predict HDF5 file. Contains:
                    dataset:    predict (either [frames, num_classes] or [frames, timesteps, num_classes])
"""
from datetime import datetime
from os import mkdir
from os.path import join
from typing import Union

import numpy as np
from docopt import docopt

import sonusai
from sonusai import create_file_handler
from sonusai import initial_log_messages
from sonusai import logger
from sonusai import update_console_handler
from sonusai.metrics import calculate_optimal_thresholds
from sonusai.metrics import class_summary
from sonusai.metrics import snr_summary
# from sonusai.metrics import one_hot
from sonusai.queries.get_organized_snrs import get_mixids_from_ordered_snr
from sonusai.utils import human_readable_size
from sonusai.utils import read_feature_data
from sonusai.utils import read_predict_data
from sonusai.utils import seconds_to_hms
from sonusai.utils import trim_docstring


def evaluate(mixdb: dict,
             truth: np.ndarray,
             predict: Union[None, np.ndarray],
             segsnr: np.ndarray = [],
             output_dir: str = None,
             pred_thr: float = 0.0,
             feature: np.ndarray = [],
             mixture: Union[None, np.ndarray] = None,
             target: Union[None, np.ndarray] = None,
             noise: Union[None, np.ndarray] = None,
             plot_number: int = 0,
             verbose: bool = False):
    update_console_handler(verbose)
    initial_log_messages('evaluate')

    if truth.shape != predict.shape:
        logger.exception(f'Shape of truth and predict are not equal. Exiting.')
        raise SystemExit(1)

    # truth, predict can be either frames x num_classes, or frames x timesteps x num_classes
    # in binary case dim may not exist, detect this and set num_classes == 1
    timesteps = -1
    if truth.ndim == 3 or (truth.ndim == 2 and timesteps > 0):
        if truth.ndim == 2:  # 2D but has timesteps = truth.shape[1]
            num_classes = 1
        else:
            num_classes = truth.shape[2]  # frames = truth.shape[0], timesteps = truth.shape[1]
        # reshape to remove timestep dimension
        truth = np.reshape(truth, (truth.shape[0] * truth.shape[1], truth.shape[2]))
        predict = np.reshape(predict, (predict.shape[0] * predict.shape[1], predict.shape[2]))
    else:
        if truth.ndim == 1:  # no timesteps dimension, = 0
            num_classes = 1
            truth = np.expand_dims(truth, 1)  # convert to 2D (F,1) if only 1 dim
        else:
            num_classes = truth.shape[1]  # 2D input frames x num_classes

    # Support predict (and truth) when shape is (F,), just convert to (F,1)
    if predict.ndim == 1:
        predict = np.expand_dims(predict, 1)

    # Check segsnr, input is always in transform frames
    compute_segsnr = False
    if len(segsnr) > 0:
        segsnr_feature_frames = segsnr.shape[0] / (mixdb['feature_step_samples'] / mixdb['frame_size'])
        if segsnr_feature_frames == truth.shape[0]:
            compute_segsnr = True
        else:
            logger.warning('segsnr length does not match truth, ignoring.')

    # Check pred_thr array or scalar and return final scalar pred_thr value
    if not mixdb['truth_mutex']:
        if num_classes > 1:
            if np.ndim(pred_thr) == 0 and pred_thr == 0:
                pred_thr = 0.5  # multi-label pred_thr scalar 0 force to 0.5 default

            if np.ndim(pred_thr) == 1:
                if len(pred_thr) == 1:
                    if pred_thr[0] == 0:
                        pred_thr = 0.5  # multi-label pred_thr array scalar 0 force to 0.5 default
                    else:
                        pred_thr = pred_thr[0]  # multi-label pred_thr array set to scalar = array[0]

    else:
        pred_thr = 0  # single-label mode, force argmax mode

    if pred_thr == -1:
        thrpr, thrroc, _, _ = calculate_optimal_thresholds(truth, predict)
        pred_thr = np.atleast_1d(thrpr)
        pred_thr = np.maximum(pred_thr, 0.1)  # enfore lower limit 0.1
        pred_thr = np.minimum(pred_thr, 0.9)  # enfore upper limit 0.9
        pred_thr = pred_thr.round(2)

    # Summarize the mixture data
    num_mixtures = len(mixdb['mixtures'])
    total_samples = sum([sub['samples'] for sub in mixdb['mixtures']])
    duration = total_samples / sonusai.mixture.SAMPLE_RATE

    logger.info('')
    logger.info(f'Mixtures: {num_mixtures}')
    logger.info(f'Duration: {seconds_to_hms(seconds=duration)}')
    logger.info(f'truth:    {human_readable_size(truth.nbytes, 1)}')
    logger.info(f'predict:  {human_readable_size(predict.nbytes, 1)}')
    if compute_segsnr:
        logger.info(f'segsnr:   {human_readable_size(segsnr.nbytes, 1)}')
    if len(feature) > 0:
        logger.info(f'feature:  {human_readable_size(feature.nbytes, 1)}')

    logger.info(f'Classes: {num_classes}')
    if mixdb['truth_mutex']:
        logger.info(f'Mode:  Single-label / truth_mutex / softmax')
    else:
        logger.info(f'Mode:  Multi-label / Binary')

    mxid_snro = get_mixids_from_ordered_snr(mixdb=mixdb)
    snrlist = list(mxid_snro.keys())
    snrlist.sort(reverse=True)
    logger.info(f'Ordered SNRs: {snrlist}\n')
    if type(pred_thr) is np.ndarray:
        logger.info(f'Prediction Threshold(s): {pred_thr.transpose()}\n')
    else:
        logger.info(f'Prediction Threshold(s): {pred_thr}\n')

    # Top-level report over all mixtures
    if compute_segsnr:
        macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb, ':', truth, predict, segsnr, pred_thr)
    else:
        macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb, ':', truth, predict, [], pred_thr)

    if num_classes > 1:
        logger.info(f'Metrics micro-avg per SNR over all {num_mixtures} mixtures:')
    else:
        logger.info(f'Metrics per SNR over all {num_mixtures} mixtures:')
    logger.info(microdf.round(3).to_string())
    logger.info('\n')
    if output_dir:
        microdf.round(3).to_csv(join(output_dir, 'snr.csv'))

    if mixdb['truth_mutex']:
        if compute_segsnr:
            macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb, ':', truth[:, 0:-1], predict[:, 0:-1], segsnr,
                                                              pred_thr)
        else:
            macrodf, microdf, wghtdf, mxid_snro = snr_summary(mixdb, ':', truth[:, 0:-1], predict[:, 0:-1], [],
                                                              pred_thr)

        logger.info(f'Metrics micro-avg without "Other" class per SNR over all {num_mixtures} mixtures:')
        logger.info(microdf.round(3).to_string())
        logger.info('\n')
        if output_dir:
            microdf.round(3).to_csv(output_dir, '/snrwo.csv')

    for snri in snrlist:
        mxids = mxid_snro[snri]
        classdf = class_summary(mixdb, mxids, truth, predict, pred_thr)
        logger.info(f'Metrics per class for SNR {snri} over {len(mxids)} mixtures:')
        logger.info(classdf.round(3).to_string())
        logger.info('\n')
        if output_dir:
            classdf.round(3).to_csv(join(output_dir, f'class_snr{snri}.csv'))

    return


#     # Number of samples per feature
#     samples_per_feature = mixdb['feature_samples']
#
#     # Total mixture audio samples
#     mixture_samples = feature_frames * samples_per_feature
#
#     # Number of samples per transform frame
#     samples_per_transform = int(mixture_samples / transform_frames)
#
#
#
#
#     if plot_number < 0:
#         # If plot_number is negative, plot all
#         plot_number = [*range(num_mixtures)]
#     elif plot_number == 0:
#         # If plot_number is zero, plot none
#         plot_number = []
#     else:
#         # plot_number needs to be a list (for looping later)
#         plot_number = [plot_number]
#
#     # Summarize mixture data by noise file, target file, SNR via array mstat:
#     # size NNFxNTFxNSNRxNAUGx20
#     #  #noise-files, #SNR, #target-files, #augmentations, #param fields.
#     # We track all other augmentations in a dimension that auto-increments, so we can support any combinations.
#     # param fields:
#     #          mi,target_snr,...
#     noise_files = mixdb['noises']
#     num_noise_files = len(noise_files)
#     target_files = mixdb['targets']
#     num_target_files = len(target_files)
#
#     # Determine number of SNR cases
#     snrs = []
#     for target_augmentation in mixdb['target_augmentations']:
#         if 'snr' in target_augmentation:
#             snrs.append(target_augmentation['snr'])
#     snrs = sorted(list(set(snrs)), reverse=True)
#
#     num_snrs = len(snrs)  # Number of SNR cases
#     # Detect special noise-only cases, specified by SNR < -96 where genft sets target gain = 0
#     np_snrs = []
#     for i, x in enumerate(snrs):
#         if float(x) < -96:
#             np_snrs.append(i)
#     # Index of noise performance SNR (should only be 1 of them)
#     if len(np_snrs) > 1:
#         logger.debug('Error: more than one noise performance SNR < -96 dB detected, proceeding with first one.')
#         np_snrs = [np_snrs[0]]
#
#     num_augmentation_types = num_mixtures / (num_noise_files * num_target_files * num_snrs)
#     if num_augmentation_types % 1:  # Number of augmentation types should always be integer
#         logger.warning('Number of augmentations is fractional.')
#     num_augmentation_types = int(np.ceil(num_augmentation_types))
#
#     logger.info(f'Detected {num_augmentation_types} augmentation types beyond noise, target, and SNR.')
#
#     if mixdb['truth_mutex']:
#         logger.debug('Detected truth mutex mode')
#         truth_is_mutex = True
#     else:
#         truth_is_mutex = False
#
#     logger.info(
#         f'Analyzing performance for {num_mixtures} mixtures, {num_noise_files} noise files, {num_target_files} target files, and {num_snrs} SNR levels.')
#
#     if binary_detection_threshold == 0:
#         binary_detection_threshold = 0.5
#         cmode = 0
#         logger.debug(
#             f'Using default argmax() for detection (binary binary_detection_threshold = {binary_detection_threshold:4.2f})')
#     else:
#         cmode = binary_detection_threshold
#         logger.debug(f'Using custom detection threshold of {binary_detection_threshold:4.2f}')
#
#     # Mixture stats
#     mstat = np.zeros(shape=(num_noise_files, num_target_files, num_snrs, num_augmentation_types, 23), dtype=np.single)
#     # Multiclass stats
#     mcmall = np.zeros(shape=(num_noise_files, num_target_files, num_snrs, num_augmentation_types, num_classes, 2, 2),
#                       dtype=np.single)
#     if num_classes > 1:
#         cmall = np.zeros(
#             shape=(num_noise_files, num_target_files, num_snrs, num_augmentation_types, num_classes, num_classes),
#             dtype=np.single)
#     else:
#         cmall = np.zeros(shape=(num_noise_files, num_target_files, num_snrs, num_augmentation_types, 2, 2),
#                          dtype=np.single)
#
#     # Keep indices in fields [gsnr,HITFA,FPR,TPR,segsnravg,mxst.segsnrpkmed,mxst.NF]
#     fnameidx = np.zeros(shape=(num_noise_files, num_target_files, num_snrs, num_augmentation_types, 5), dtype=np.int16)
#
#     # Initial augmentation type
#     augi = 0
#     sub_predict = []
#
#     miwidth = len(str(num_mixtures - 1))
#     tfwidth = max(len(ele) for ele in [Path(x['name']).stem for x in target_files])
#     nfwidth = max(len(ele) for ele in [Path(x['name']).stem for x in noise_files])
#
#     if plot_number:
#         pdf = PdfPages(f'{output_dir}/evaluate.pdf')
#
#     for mi in range(num_mixtures):
#         ni = mixdb['mixtures'][mi]['noise_file_index']
#         ti = mixdb['mixtures'][mi]['target_file_index']
#
#         # Get mixture config parameters target gain (and target SNR above)
#         target_augmentation_index = mixdb['mixtures'][mi]['target_augmentation_index']
#         target_snr = mixdb['target_augmentations'][target_augmentation_index]['snr']
#
#         if 'gain' in mixdb['target_augmentations'][target_augmentation_index]:
#             target_gain = mixdb['target_augmentations'][target_augmentation_index]['gain']
#         else:
#             target_gain = 1
#
#         target_snr_index = snrs.index(target_snr)
#
#         # record the FIRST truth index, provided by genft per mixture. (TBD - change to support multiple tidx??)
#         truth_index = mixdb['targets'][mixdb['mixtures'][mi]['target_file_index']]['truth_index'][0][0] - 1
#         # store index into filename lists in case we want full filenames
#         fnameidx[ni, ti, target_snr_index, augi,] = [ni, ti, truth_index, target_snr_index, augi]
#
#         # For each mixture, get index endpoints (mix,noise,target sample,...)
#         i_sample_begin = mixdb['mixtures'][mi]['i_sample_offset']
#         i_frame_begin = mixdb['mixtures'][mi]['i_frame_offset']
#         o_frame_begin = mixdb['mixtures'][mi]['o_frame_offset']
#         if mi == num_mixtures - 1:
#             i_sample_end = mixture_samples
#             i_frame_end = transform_frames
#             o_frame_end = feature_frames
#         else:
#             i_sample_end = mixdb['mixtures'][mi + 1]['i_sample_offset']
#             # in frames are transform
#             i_frame_end = mixdb['mixtures'][mi + 1]['i_frame_offset']
#             # out frames are possibly decimated/stride-combined
#             o_frame_end = mixdb['mixtures'][mi + 1]['o_frame_offset']
#
#         # Select subset of waveforms, if present
#         sub_mixture = []
#         if mixture is not None:
#             sub_mixture = int16_to_float(mixture[i_sample_begin:i_sample_end])
#         sub_noise = []
#         if noise is not None:
#             sub_noise = int16_to_float(noise[i_sample_begin:i_sample_end])
#         sub_target = []
#         if target is not None:
#             sub_target = int16_to_float(target[i_sample_begin:i_sample_end])
#
#         noise_level = get_level(sub_noise)
#         target_level = get_level(sub_target)
#
#         mixture_snr = target_level - noise_level
#
#         # Segmental SNR and target stats
#         sub_segnsr = segsnr[i_frame_begin:i_frame_end]
#
#         # Replace float('Inf') with weighted filter of 16 past samples, then median filter
#         sub_segnsr = segsnr_filter(sub_segnsr)
#
#         # segmental SNR mean = mixture_snr and target_snr
#         ssnrmean = 10 * np.log10(np.mean(sub_segnsr + 1e-10))
#         # seg SNR max
#         ssnrmax = 10 * np.log10(max(sub_segnsr + 1e-10))
#         # seg SNR 80% percentile
#         ssnrpk80 = 10 * np.log10(np.percentile(sub_segnsr + 1e-10, 80, interpolation='midpoint'))
#
#         # Truth and neural-net prediction data, feature_frames x num_classes
#         sub_truth = truth[o_frame_begin:o_frame_end, ]
#
#         if predict is not None:
#             sub_predict = predict[o_frame_begin:o_frame_end]
#             # don't do this, expect genft to do the right thing w/truth, must be mutex
#             # special case for false alarm analysis, set truth to zeros
#             # if float(target_snr) < -96:
#             #     truth = np.zeros(truth.shape)
#
#             # ACC TPR PPV TNR FPR HITFA F1 MCC NT PT for each class, nclass x 10
#             # cm is [TN FP; FN TP]
#             (mcm, metrics, cm, cmn, rmse) = one_hot(sub_truth, sub_predict, cmode)
#             # mcm2, metrics2, cm2, cmn2, rmse2 = classmetrics(sub_truth, sub_predict, cmode)
#
#             # Save all stats, metrics and TN FP FN TP, but only includes for truth_index class.
#             mstat[ni, ti, target_snr_index, augi,] = [mi, target_snr, ssnrmean, ssnrmax, ssnrpk80, target_gain,
#                                                       *metrics[truth_index,],
#                                                       mcm[truth_index, 0, 0], mcm[truth_index, 0, 1],
#                                                       mcm[truth_index, 1, 0],
#                                                       mcm[truth_index, 1, 1],
#                                                       rmse[truth_index]]
#             # metricsall[ni, ti, target_snr_index, augi, ] = metrics
#             mcmall[ni, ti, target_snr_index, augi,] = mcm
#             cmall[ni, ti, target_snr_index, augi,] = cm
#             # cmnall[ni, ti, target_snr_index, augi, ] = cmn
#
#             # Increment augmentation type
#             augi += 1
#             if augi >= num_augmentation_types:
#                 augi = 0
#
#         if mi + 1 in plot_number:
#             tname = Path(target_files[ti]['name']).stem
#             nname = Path(noise_files[ni]['name']).stem
#             logger.debug(
#                 f'{mi:{miwidth}}: Mix spec: SNR: {float(target_snr):4.1f}, target gain: {target_gain:+4.1f}, file: {tname:>{tfwidth}}, {nname:>{nfwidth}}')
#             logger.debug(
#                 f'{mi:{miwidth}}: Measured: SNR: {mixture_snr:4.1f}, target level: {target_level:4.1f}, SegSNR mean, max, peak80%: {ssnrmean:4.1f}, {ssnrmax:4.1f}, {ssnrpk80:4.1f}')
#
#             # Expand number of frames to number of samples (for plotting together with waveforms)
#             segsnrex = np.reshape(np.tile(10 * np.log10(sub_segnsr + np.finfo(float).eps), [samples_per_transform, 1]),
#                                   len(sub_segnsr) * samples_per_transform,
#                                   order='F')
#             sub_feature = feature[o_frame_begin:o_frame_end, ]
#
#             # Truth extension to #samples in this mixture, i.e. MixSubSamples x num_classes
#             truthex = np.zeros((len(sub_segnsr) * samples_per_transform, num_classes))
#             for ntri in range(num_classes):
#                 truthex[:, ntri] = np.reshape(np.tile(sub_truth[:, ntri], [samples_per_feature, 1]),
#                                               len(sub_segnsr) * samples_per_transform, order='F')
#
#             # vector of indices where truth is present
#             truth_active = []
#             for i, x in enumerate(np.any(sub_truth, axis=0)):
#                 if x:
#                     truth_active.append(i)
#
#             if truth_is_mutex:
#                 # remove mutex (end), but not when class is always active
#                 if len(truth_active) > 1:
#                     truth_active = np.delete(truth_active, -1)
#
#             if len(truth_active) != 0:
#                 logger.debug(
#                     f'{mi:{miwidth}}: Truth present in {len(truth_active)} of {num_classes} classes. First active category index: {truth_active[0]}')
#             else:
#                 logger.debug(f'{mi:{miwidth}}: Truth not present in {num_classes} classes.')
#
#             fig, (ax1, ax2, ax3) = plt.subplots(3, 1)
#
#             secu = np.arange(len(sub_segnsr) * samples_per_transform, dtype=np.single) / sonusai.mixture.SAMPLE_RATE
#             plots = []
#             ymin = [0]
#
#             if mixture is not None:
#                 mix_plot, = ax1.plot(secu, sub_mixture, 'r', label='Mix')
#                 plots.append(mix_plot)
#                 ymin.append(min(sub_mixture))
#             if target is not None:
#                 target_plot, = ax1.plot(secu, sub_target, 'b', label='Target')
#                 plots.append(target_plot)
#                 ymin.append(min(sub_target))
#
#             ax1.set_ylim([min(ymin), 1.05])
#
#             if len(truth_active) > 0:
#                 if len(truth_active) == 1:
#                     truthn_plot, = ax1.plot(secu, truthex[:, truth_active[0]], 'g-',
#                                             label=f'NN Truth{truth_active[0]}')
#                     plots.append(truthn_plot)
#                     if truth_is_mutex:
#                         trutho_plot, = ax1.plot(secu, truthex[:, -1], 'm-.',
#                                                 label=f'NN Truth{num_classes} (Other)')
#                         plots.append(trutho_plot)
#                 else:
#                     trutha_plot, = ax1.plot(secu, truthex, 'g-', label='NN Truth All')
#                     plots.append(trutha_plot)
#
#             if sub_predict is not None:
#                 nnoex = np.reshape(np.tile(sub_predict, [samples_per_feature, 1]),
#                                    len(sub_segnsr) * samples_per_transform)
#                 nno_plot, = ax1.plot(secu, nnoex, 'k', label='NN Predict All')
#                 plots.append(nno_plot)
#
#             ax1.legend(handles=plots, bbox_to_anchor=(1.05, 1.0), loc='upper left')
#
#             ax2.plot(secu, segsnrex)
#             ax2.set_ylim([-51, max(min(max(segsnrex), 100), -10)])
#             ax2.title.set_text(
#                 f'SegSNR mean={ssnrmean:4.1f} dB, max={ssnrmax:4.1f} dB, Peak80%={ssnrpk80:4.1f} dB')
#
#             if sub_feature.shape[1] < 5:
#                 tmp = np.reshape(sub_feature, (sub_feature.shape[0], sub_feature.shape[1] * sub_feature.shape[2]))
#                 titlestr = f'All {sub_feature.shape[0]} features of size {sub_feature.shape[2]}x{sub_feature.shape[1]}'
#             else:
#                 if len(truth_active) != 0:
#                     fti = np.nonzero(sub_truth[:, truth_active[0]])[0][0]
#                 else:
#                     fti = 0
#                 tmp = np.squeeze(sub_feature[fti,])
#                 titlestr = f'Example feature, frame {fti} of size {sub_feature.shape[2]}x{sub_feature.shape[1]}'
#             ax3.imshow(np.transpose(tmp), aspect='auto')
#             ax3.invert_yaxis()
#             ax3.title.set_text(titlestr)
#
#             fig.suptitle(f'{mi + 1} of {num_mixtures}')
#             fig.tight_layout()
#             pdf.savefig(fig)
#             plt.close(fig)
#
#     if plot_number:
#         pdf.close()
#
#     if predict is not None:
#         save_eval_report(output_dir=output_dir,
#                          noise_files=noise_files,
#                          target_files=target_files,
#                          snrs=snrs,
#                          fnameidx=fnameidx,
#                          binary_detection_threshold=binary_detection_threshold,
#                          np_snrs=np_snrs,
#                          truth_is_mutex=truth_is_mutex,
#                          mstat=mstat,
#                          mcmall=mcmall,
#                          cmall=cmall)
#
#
# def median_filter(x: np.ndarray, n: int) -> np.ndarray:
#     assert x.ndim == 1, 'Input must be one-dimensional.'
#
#     n2 = (n - 1) // 2
#     if n % 2 == 0:
#         n2 += 1
#
#     pad_x = np.zeros(len(x) + n, dtype=x.dtype)
#     pad_x[n2:n2 + len(x)] = x
#
#     y = np.zeros((len(x), n), dtype=x.dtype)
#     for k in range(len(x)):
#         y[k,] = pad_x[k:k + n]
#
#     return np.median(y, axis=1)
#
#
# def segsnr_filter(segsnr: np.ndarray) -> np.ndarray:
#     assert segsnr.ndim == 1, 'Input must be one-dimensional.'
#
#     inf_indices = [i for i, val in enumerate(segsnr) if np.isinf(val)]
#
#     if inf_indices:
#         scale = np.arange(16) + 1
#         for inf_index in inf_indices:
#             # avoid underflow at beginning
#             if inf_index > 15:
#                 i = range(inf_index - 1, inf_index - 17, -1)
#                 segsnr[inf_index] = np.sum(np.multiply(segsnr[i], scale)) / 136
#             else:
#                 i = range(inf_index + 1)
#                 segsnr[inf_index] = np.median(segsnr[i])
#         segsnr = median_filter(segsnr, 32)
#     return segsnr
#
#
# def save_eval_report(output_dir: str,
#                      noise_files: List[Dict],
#                      target_files: List[Dict],
#                      snrs: dict,
#                      fnameidx: np.ndarray,
#                      binary_detection_threshold: float,
#                      np_snrs: list,
#                      truth_is_mutex: bool,
#                      mstat: np.ndarray,
#                      mcmall: np.ndarray,
#                      cmall: np.ndarray,
#                      nfacrit: float = 0.05):
#     outname = 'evaluate'
#     num_noise_files = len(noise_files)
#     num_target_files = len(target_files)
#     num_snrs = len(snrs)
#     num_augmentation_types = mstat.shape[3]
#     NCLASS = mcmall.shape[4]  # mcm will = 1 for binary (always be correct)
#     NMIX = num_noise_files * num_target_files * num_snrs * num_augmentation_types
#     logger.info(
#         f'Printing reports in {output_dir} for {NMIX} mixtures, {num_noise_files} noise files, {num_target_files} target files, {num_snrs} SNRs, {num_augmentation_types} augmentations')
#
#     # Target all SNR list index, ordered from highest to lowest SNR
#     tasnridx = np.flip(np.argsort(mstat[0, 0, :, 0, 1]))
#     # Detect special case SNR
#     if len(np_snrs) == 0:
#         np_snrs = len(snrs) - 1
#         logger.warning(
#             f'Did not find low SNR (<-96 dB) case, FA analysis uses lowest SNR of {float(snrs[np_snrs]):3.0f}\n')
#     else:
#         np_snrs = np_snrs[0]  # In case a list is returned, use 0, TBD list support
#
#     FPSNR = float(snrs[np_snrs])
#     # Create list of non-special case SNRs, for some metrics
#     if FPSNR < -96:
#         sridx = np.arange(0, num_snrs)
#         sridx = np.delete(sridx, np_snrs)  # Remove special case SNR
#         srsnrlist = list(map(snrs.__getitem__, sridx))  # new list wo fpsnr case
#     else:
#         FPSNR = 99  # Mark as unused with value 99
#         sridx = tasnridx
#         srsnrlist = snrs
#
#     snrfloat = np.zeros((num_snrs, 1))
#     for i, x in enumerate(snrs):
#         snrfloat[i] = float(x)
#
#     if NCLASS == 1:
#         bmode = True
#     else:
#         bmode = False  # TBD might need to detect NCLASS == 2 and truth_is_mutex
#
#     # Total metrics over all data
#     metsa, mcmsa, cmsumsa = calculate_metrics(cmall, bmode)  # metric summary all
#     metsa_avg = averages(metsa)  # 3 rows: [PPV, TPR, F1, FPR, ACC, TPSUM]
#     metsa_avgwo = averages(metsa[0:-1, ])
#
#     # Total metrics over all non-special case SNR
#     metsn, mcmsn, cmsumsn = calculate_metrics(cmall[:, :, sridx, ], bmode)
#     metsn_avg = averages(metsn)  # 3x6
#     metsn_avgwo = averages(metsn[0:-1, ])  # without other (last) class
#
#     # Metric summary by noise and snr per class:
#     # fields: [ACC, TPR, PPV, TNR, FPR, HITFA, F1, MCC, NT, PT, TP, FP]
#     metsnr = np.zeros((num_snrs, NCLASS, 12))  # per SNR: nsnr x class x 12
#     metsnr_avg = np.zeros((num_snrs, 3, 6))  # per SNR avg over classes: nsnr x 3 x 6
#     metsnr_avgwo = np.zeros((num_snrs, 3, 6))  # without other class
#     metnsnr = np.zeros((num_noise_files, num_snrs, NCLASS, 12))  # Per Noise & SNR: nnf x nsnr x nclass x 12
#     metnsnr_avg = np.zeros((num_noise_files, num_snrs, 3, 6))
#     metnsnr_avgwo = np.zeros((num_noise_files, num_snrs, 3, 6))
#     for target_snr_index in range(num_snrs):
#         metsnr[target_snr_index,], _, _ = calculate_metrics(cmall[:, :, target_snr_index, ])
#         metsnr_avg[target_snr_index,] = averages(metsnr[target_snr_index,])
#         metsnr_avgwo[target_snr_index,] = averages(metsnr[target_snr_index, 0:-1, ])  # without other class = nclass
#         for ni in range(num_noise_files):
#             metnsnr[ni, target_snr_index,], _, _ = calculate_metrics(cmall[ni, :, target_snr_index, ])
#             metnsnr_avg[ni, target_snr_index,] = averages(metnsnr[ni, target_snr_index,])
#             metnsnr_avgwo[ni, target_snr_index,] = averages(metnsnr[ni, target_snr_index, 0:-1, ])
#
#     # mcm: [TN FP; FN TP]
#     # Fields in mstat, TargetSNR, TargetGain:
#     FTSNR = 1
#     FTGAIN = 5
#     # Fields in metrics:
#     FACC = 0
#     FF1 = 6
#     FTPR = 1
#     FFA = 4
#     FHITFA = 5
#
#     # FA per noise-file @ lowest SNR, averages over classes:
#     fanf_macwo = metnsnr_avgwo[:, np_snrs, 0, 3]  # macro-average wo other class NNFx1
#     fanf_micwo = metnsnr_avgwo[:, np_snrs, 1, 3]  # micro-average wo other class NNFx1
#     fanf_mac = metnsnr_avg[:, np_snrs, 0, 3]  # macro-average NNFx1
#     fanf_mic = metnsnr_avg[:, np_snrs, 1, 3]  # micro-average NNFx1
#     nfafail = [i for i, x in enumerate(fanf_mic) if x > nfacrit]
#
#     # Create summary report
#     summary_name = f'{output_dir}/nn_perfsummary.txt'
#     with open(file=summary_name, mode='w') as f:
#         # RMSEmean = np.squeeze(np.mean(np.reshape(mstat[:, :, :, :, 20], (num_noise_files, num_target_files * num_augmentation_types * num_snrs, 1)), 1))
#
#         if bmode:  # -------------- Binary Classification ------------------
#             f.write('\n')
#             f.write(f'--- Aaware Binary NN Performance Analysis: {outname}\n')
#             f.write(f'Performance over {NMIX} mixtures at {num_snrs} SNR cases created from\n')
#             f.write(
#                 f'{num_target_files} target files and {num_noise_files} noise files and {num_augmentation_types} augmentations.\n')
#             f.write(f'Binary classification threshold: {binary_detection_threshold:4.2f}.\n')
#             f.write('\n')
#             # FA Summary, note: for binary *_avg are accurate, although no avg needed (but avgwo is not)
#             f.write(f'--- False Alarm Performance over {num_noise_files} noise files  ---\n')
#             f.write(f'FA over all mixtures:             {metsa_avg[1, 3] * 100:6.5f}%\n')
#             f.write(f'Max FA over SNRs and Noise files: {np.max(metnsnr_avg[:, :, 1, 3]) * 100:6.5f}%\n')
#             f.write(f'FA of lowest SNR {FPSNR:2.0f} dB:          {metsnr_avg[np_snrs, 1, 3] * 100:6.5f}%\n')
#             if len(nfafail) > 0:
#                 f.write(f'FAIL: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)}\n')
#             else:
#                 f.write(f'PASS: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)}\n')
#
#             f.write(generate_snr_summary(mstat, metsnr, snrfloat, tasnridx))
#
#             f.write('\n\n')
#             # Failing Noise File List
#             if len(nfafail) > 0:
#                 f.write(
#                     f'--- List of {len(nfafail)} failing noise files with FA rate > {nfacrit * 100:3.1f}%:\n')
#                 for jj in range(len(nfafail)):
#                     f.write(f'{noise_files[nfafail[jj]]["name"]:40s}: {100 * fanf_mic[nfafail[jj]]:3.1f}%\n')
#
#             f.write('\n')
#
#         else:  # ----------------- Multiclass Classification -----------------------------
#             f.write('\n')
#             f.write(f'--- Aaware Multiclass NN Performance Analysis: {outname}\n')
#             f.write(f'Performance over {NMIX} mixtures at {num_snrs} SNR cases created from\n')
#             f.write(
#                 f'{num_target_files} target files and {num_noise_files} noise files and {num_augmentation_types} augmentations.\n')
#             f.write(f'Number of classes: {NCLASS}.\n')
#             f.write(f'Mutex mode (single-label): {truth_is_mutex}\n')
#             f.write(f'Classification decision threshold: {binary_detection_threshold:4.2f}.\n')
#             f.write('Confusion Matrix over all mixtures:\n')
#             f.write(generate_snr_summary(mstat, metsnr, snrfloat, tasnridx))
#             f.write('\n')
#             # Metrics over all mixtures:
#             f.write('Confusion Matrix over all mixtures:\n')
#             f.write(format_confusion_matrix(cmsumsa))
#             f.write('Metric summary, all mixtures:\n')
#             f.write(generate_summary(metsa))
#             f.write('\n')
#             # Metrics excluding special case SNR
#             if FPSNR < -96:
#                 f.write(
#                     f'Metrics excluding special case false-positive SNR (FPSNR) at {FPSNR:2.0f} dB\n')
#                 f.write('Confusion Matrix for all mixtures except FPSNR:\n')
#                 f.write(format_confusion_matrix(cmsumsn))
#
#             # FA Summary
#             f.write(
#                 f'False alarm over all mixtures, class micro-avg for all, non-other: {metsa_avg[1, 3] * 100:6.5f}%, {metsa_avgwo[1, 3] * 100:6.5f}%\n')
#             f.write(
#                 f'False alarm summary for lowest SNR at {FPSNR:2.0f} dB, class micro-avg over all, non-other: {metsnr_avg[np_snrs, 1, 3] * 100:6.5f}%, {metsnr_avgwo[np_snrs, 1, 3] * 100:6.5f}%\n')
#             if len(nfafail) > 0:
#                 f.write(
#                     f'FAIL: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)} (see noise-breakdown report)\n')
#             else:
#                 f.write(f'PASS: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)}\n')
#             # Print false alarm summary, use micro-avg which tends more pessimistic excluding other class if in mutex mode
#
#             f.write('\n\n')
#             f.write('-------- Per SNR Metrics ------------ \n')
#             f.write('\n')
#             for target_snr_index in range(num_snrs):
#                 si = tasnridx[target_snr_index]  # sorted list
#                 f.write('\n')
#                 f.write(f'--- Confusion Matrix for SNR {float(snrs[si]):3.1f} dB:\n')
#                 (met, _, cmsum) = calculate_metrics(cmall[:, :, si, ])
#                 f.write(format_confusion_matrix(cmsum))
#                 f.write('Metric summary:\n')
#                 f.write(generate_summary(metsnr[si,]))
#                 f.write('\n\n')
#
#     # --------------- Create noise breakdown report --------------------------------
#     noise_breakdown_name = f'{output_dir}/noise_breakdown.txt'
#     with open(file=noise_breakdown_name, mode='w') as f:
#         # Calc #frames per noise file, all speech,augm cases at lowSNR case
#         f.write('\n')
#         f.write(f'--- Aaware sonusai() NN False Alarm Analysis: {outname}\n')
#         f.write(f'Performance over {NMIX} mixtures at {num_snrs} SNR cases created from\n')
#         f.write(
#             f'{num_target_files} target files, {num_noise_files} noise files and {num_augmentation_types} augmentations.\n')
#         f.write(f'Classification decision threshold: {binary_detection_threshold:4.2f}.\n')
#         f.write(f'Number of classes: {NCLASS}.\n')
#         f.write(f'Mutex mode (single-label): {truth_is_mutex}\n')
#         f.write('\n')
#         f.write(f'--- NN False Alarm Summary over {num_noise_files} noise files ---\n')
#         # FA Summary
#         # Print false alarm summary, use micro-avg which tends more pessimistic excluding other class if in mutex mode
#         f.write(
#             f'False alarm over all mixtures, class micro-avg for all, non-other: {metsa_avg[1, 3] * 100:6.5f}%,{metsa_avgwo[1, 3] * 100:6.5f}%\n')
#         f.write(
#             f'False alarm summary for lowest SNR at {FPSNR:2.0f} dB, class micro-avg over all, non-other: {metsnr_avg[np_snrs, 1, 3] * 100:6.5f}%, {metsnr_avgwo[np_snrs, 1, 3] * 100:6.5f}%\n')
#
#         if len(nfafail) > 0:
#             f.write(f'FAIL: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)}\n')
#         else:
#             f.write(f'PASS: Number of failing noise files with FPR rate > {nfacrit * 100:3.1f}%: {len(nfafail)}\n')
#         # FA per SNR
#         f.write('False alarm per SNR micro-avg (%):\n')
#         f.write(f'{np.transpose(snrfloat[tasnridx].astype(int))}\n')
#         f.write(f'{np.round(metsnr_avg[tasnridx, 1, 3] * 100, 2)}\n')
#         # FA per SNR and Class:
#         # f.write(f'False alarm per SNR and Class (%):\n {np.round(metsnr[:,:,FFA]*100,2)}\n')
#
#         f.write('\n')
#         NFLOWSNR = int(
#             np.sum((metnsnr[:, np_snrs, 0, 8] + metnsnr[:, np_snrs, 0, 9])) / num_noise_files)  # Nframes = NT + PT
#         f.write(
#             f'--- FA perf. per noise file, over {num_target_files * num_augmentation_types} mixtures, {NFLOWSNR} frames per SNR:\n')
#
#         nfileidx = fnameidx[:, 0, 0, 0, 0]
#         for ni in range(num_noise_files):
#             # sortni = issnrl[ni]   # TBD figure out how to sort by FA perf
#             # FA per SNR
#             tmpnfname = splitext(basename(noise_files[nfileidx[ni]]['name']))[0]
#             f.write('\n')
#             f.write(f'--- NN Perf. for noise file {ni + 1} of {num_noise_files}: {tmpnfname} ---\n')
#             f.write('False alarm per SNR micro-avg (%):\n')
#             f.write(f'{np.transpose(snrfloat[tasnridx].astype(int))}\n')
#             f.write(f'{np.round(metnsnr_avg[ni, tasnridx, 1, 3] * 100, 2)}\n')
#             f.write('Confusion Matrix:\n')
#             (_, _, cmsum) = calculate_metrics(cmall[ni,])
#             f.write(format_confusion_matrix(cmsum))
#             f.write(f'Noise file full path: {noise_files[nfileidx[ni]]["name"]}\n')
#
#
# def get_level(audio: np.ndarray) -> float:
#     if audio is not None:
#         # level in dB
#         return 20 * np.log10(np.sqrt(np.mean(np.square(audio))) + np.finfo(float).eps)
#     return float('nan')


def main():
    try:
        args = docopt(trim_docstring(__doc__), version=sonusai.__version__, options_first=True)

        feature_name = args['--feature']

        if args['--predict']:
            predict_name = args['--predict']
        else:
            predict_name = ''

        predict_threshold = float(args['--thr'])

        plot_number = int(args['--plotnum'])

        # create output directory
        output_dir = f'evaluate-{datetime.now():%Y%m%d}'
        try:
            mkdir(output_dir)
        except OSError as error:
            output_dir = f'evaluate-{datetime.now():%Y%m%d-%H%M%S}'
            try:
                mkdir(output_dir)
            except OSError as error:
                logger.error(f'Could not create directory, {output_dir}: {error}')
                raise SystemExit(1)

        log_name = output_dir + '/evaluate.log'
        create_file_handler(log_name)

        mixdb, feature, truth, segsnr = read_feature_data(feature_name)
        # mixture, target, noise = read_mixture_data(mixture_name)
        predict = read_predict_data(predict_name, feature.shape[0])

        evaluate(mixdb=mixdb,
                 truth=truth,
                 segsnr=segsnr,
                 output_dir=output_dir,
                 predict=predict,
                 pred_thr=predict_threshold,
                 plot_number=plot_number,
                 verbose=args['--verbose'])

        logger.info(f'Wrote results to {output_dir}')

    except KeyboardInterrupt:
        logger.info('Canceled due to keyboard interrupt')
        raise SystemExit(0)


if __name__ == '__main__':
    main()



def update_config_from_file(name: str, config: dict) -> dict:
    from copy import deepcopy

    from sonusai import SonusAIError
    from sonusai.mixture.load_config import load_config
    from sonusai.mixture.update_truth_settings import update_truth_settings

    new_config = deepcopy(config)

    try:
        given_config = load_config(name)
    except Exception as e:
        raise SonusAIError(f'Error loading config from {name}: {e}')

    # Use default config as base and overwrite with given config keys as found
    for key in new_config:
        if key in given_config:
            if key != 'truth_settings':
                new_config[key] = given_config[key]

    if 'truth_settings' in given_config:
        new_config['truth_settings'] = deepcopy(given_config['truth_settings'])

    if not isinstance(new_config['truth_settings'], list):
        new_config['truth_settings'] = [new_config['truth_settings']]

    default_ts = deepcopy(config['truth_settings'])
    if not isinstance(default_ts, list):
        default_ts = [default_ts]

    update_truth_settings(new_config['truth_settings'], default_ts)

    required_keys = [
        'class_labels',
        'class_weights_threshold',
        'feature',
        'frame_size',
        'noises',
        'noise_augmentations',
        'num_classes',
        'seed',
        'target_augmentations',
        'targets',
        'truth_settings',
        'truth_mode',
        'truth_reduction_function',
    ]
    for key in required_keys:
        if key not in new_config:
            raise SonusAIError(f"Missing required '{key}' in {name}")

    return new_config



from typing import Any
from typing import Callable
from typing import List
from typing import Union

import numpy as np


def _process_snrs(mixdb: dict) -> (int, np.ndarray, np.ndarray):
    snrs_dict = dict()
    for index, augmentation in enumerate(mixdb['target_augmentations']):
        snr = augmentation['snr']
        if snr not in snrs_dict:
            snrs_dict[snr] = list()
        snrs_dict[snr].append(index)
    snrs_list = np.array(list(snrs_dict))
    count = np.zeros(len(snrs_dict), 'int32')
    for sni in range(len(snrs_dict)):
        snr = snrs_list[sni]
        count[sni] = len(snrs_dict[snr])
    ssidx = np.argsort(-count)
    ss = count[ssidx]
    snr_maxcnt = ss[0]
    ord_len = 1
    ii = 1
    while ii < len(snrs_dict) and ss[ii] == snr_maxcnt:
        ord_len += 1
        ii += 1
    return ord_len, snrs_list, ssidx


def get_random_snrs(mixdb: dict, predicate: Callable[[Any], bool] = None) -> list:
    ord_len, snrs_list, ssidx = _process_snrs(mixdb)

    if predicate is None:
        def predicate(x: Any) -> bool:
            return True

    snrs = sorted(list(snrs_list[ssidx[ord_len:]]), reverse=True)

    # Filter on predicate
    snrs = [snr for snr in snrs if predicate(snr)]

    return snrs


def get_ordered_snrs(mixdb: dict, predicate: Callable[[Any], bool] = None) -> list:
    ord_len, snrs_list, ssidx = _process_snrs(mixdb)

    if predicate is None:
        def predicate(x: Any) -> bool:
            return True

    snrs = sorted(list(snrs_list[ssidx[0:ord_len]]), reverse=True)

    # Filter on predicate
    snrs = [snr for snr in snrs if predicate(snr)]

    return snrs


def get_mixids_from_ordered_snr(mixdb: dict,
                                mixid: Union[str, List[int]] = None,
                                predicate: Callable[[Any], bool] = None) -> dict:
    """
    Generate mixids based on an SNR predicate for ordered SNRs
    Return a dictionary where:
        - keys are the SNRs
        - values are lists of the mixids that match the SNR
    """
    from sonusai.queries.queries import get_mixids_from_snr

    snrs = get_ordered_snrs(mixdb)

    if predicate is None:
        def _predicate(x: Any) -> bool:
            return x in snrs
    else:
        def _predicate(x: Any) -> bool:
            return predicate(x) and x in snrs

    return get_mixids_from_snr(mixdb=mixdb, mixid=mixid, predicate=_predicate)


def get_mixids_from_random_snr(mixdb: dict,
                               mixid: Union[str, List[int]] = None,
                               predicate: Callable[[Any], bool] = None) -> dict:
    """
    Generate mixids based on an SNR predicate for random SNRs
    Return a dictionary where:
        - keys are the SNRs
        - values are lists of the mixids that match the SNR
    """
    from sonusai.queries.queries import get_mixids_from_snr

    snrs = get_random_snrs(mixdb)

    if predicate is None:
        def _predicate(x: Any) -> bool:
            return x in snrs
    else:
        def _predicate(x: Any) -> bool:
            return predicate(x) and x in snrs

    return get_mixids_from_snr(mixdb=mixdb, mixid=mixid, predicate=_predicate)



def get_hierarchical_config_files(root: str, leaf: str) -> list:
    import os
    from pathlib import Path

    from sonusai import SonusAIError

    config_file = 'config.yml'

    root_path = Path(os.path.abspath(root))
    if not root_path.is_dir():
        raise SonusAIError(f'Given root, {root_path}, is not a directory.')

    leaf_path = Path(os.path.abspath(leaf))
    if not leaf_path.is_dir():
        raise SonusAIError(f'Given leaf, {leaf_path}, is not a directory.')

    common = os.path.commonpath((root_path, leaf_path))
    if os.path.normpath(common) != os.path.normpath(root_path):
        raise SonusAIError(f'Given leaf, {leaf_path}, is not in the hierarchy of the given root, {root_path}')

    top_config_file = Path(os.path.join(root_path, config_file))
    if not top_config_file.is_file():
        raise SonusAIError(f'Could not find {top_config_file}')

    current = leaf_path
    config_files = list()
    while current != root_path:
        local_config_file = Path(os.path.join(current, config_file))
        if local_config_file.is_file():
            config_files.append(local_config_file)
        current = current.parent

    config_files.append(top_config_file)
    return list(reversed(config_files))


def update_config_from_hierarchy(root: str, leaf: str, config: dict) -> dict:
    from copy import deepcopy

    from sonusai.mixture.update_config_from_file import update_config_from_file

    new_config = deepcopy(config)
    config_files = get_hierarchical_config_files(root=root, leaf=leaf)
    for config_file in config_files:
        new_config = update_config_from_file(name=config_file, config=new_config)

    return new_config



import numpy as np


def build_noise_audio_db(mixdb: dict, show_progress: bool = False) -> list:
    from tqdm import tqdm

    from sonusai.mixture.apply_augmentation import apply_augmentation
    from sonusai.mixture.read_audio import read_audio

    db = list()
    for file_index in tqdm(sorted(list(set([sub['noise_file_index'] for sub in mixdb['mixtures']]))),
                           desc='Read noise audio', disable=not show_progress):
        audio_in = read_audio(name=mixdb['noises'][file_index]['name'])
        for augmentation_index in sorted(list(set([sub['noise_augmentation_index'] for sub in mixdb['mixtures']]))):
            db.append({
                'file':         file_index,
                'augmentation': augmentation_index,
                'audio':        apply_augmentation(audio_in=audio_in,
                                                   augmentation=mixdb['noise_augmentations'][augmentation_index],
                                                   length_common_denominator=1)
            })
    return db


def get_noise_audio_from_db(db: list, file_index: int, augmentation_index: int) -> np.ndarray:
    from sonusai import SonusAIError

    for record in db:
        if record['file'] == file_index and record['augmentation'] == augmentation_index:
            return record['audio']

    raise SonusAIError(f'Could not find file_index {file_index} and augmentation_index '
                       f'{augmentation_index} in noise database')


_audio_db_global = dict()


def build_target_audio_db(mixdb: dict, show_progress: bool = False) -> list:
    from tqdm import tqdm

    from sonusai.utils import p_map

    _audio_db_global['mixdb'] = mixdb
    indices = sorted(list(set([sub['target_file_index'] for sub in mixdb['mixtures']])))
    progress = tqdm(total=len(indices), desc='Read target audio', disable=not show_progress)
    db = p_map(_read_target_audio, indices, progress=progress)
    return db


def _read_target_audio(file_index: int) -> dict:
    from sonusai.mixture.read_audio import read_audio

    return {
        'file':  file_index,
        'audio': read_audio(name=_audio_db_global['mixdb']['targets'][file_index]['name'])
    }


def get_target_audio_from_db(db: list, file_index: int) -> np.ndarray:
    from sonusai import SonusAIError

    for record in db:
        if record['file'] == file_index:
            return record['audio']

    raise SonusAIError(f'Could not find file_index {file_index} in target database')


def check_audio_files_exist(mixdb: dict) -> None:
    from os.path import exists
    from os.path import expandvars

    from sonusai import SonusAIError

    for file_index in sorted(list(set(sub['noise_file_index'] for sub in mixdb['mixtures']))):
        file_name = expandvars(mixdb['noises'][file_index]['name'])
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')

    for file_index in sorted(list(set(sub['target_file_index'] for sub in mixdb['mixtures']))):
        file_name = expandvars(mixdb['targets'][file_index]['name'])
        if not exists(file_name):
            raise SonusAIError(f'Could not find {file_name}')
